import logging
import optparse
import os
import re
import subprocess
import sys

from noblogsmv import state
from noblogsmv import webapp


BLOGS_DIR = '/opt/noblogs/www/wp-content/blogs.dir'
MYSQL_URI_RE = re.compile(
    r'^(\d+)\s+\S+\s+mysql://([^:]+):([^@]+)@([^:]+):(\d+)/(\S+)')

log = logging.getLogger(__name__)


def init(blog_id, value, progress):
    return 'move_data'


def process_move_data(blog_id, value, progress):
    progress.update('moving blog data')
    log.info('moving data for %s', blog_id)
    blog_dir = os.path.join(BLOGS_DIR, blog_id)
    cmd = ['ssh', 'root@%s' % value['old_host'], 'rsync', '-az',
           '%s/' % blog_dir, 'root@%s:%s' % (value['new_host'], blog_dir)]
    log.debug('Calling %s', ' '.join(cmd))
    try:
        subprocess.check_call(cmd)
    except (subprocess.CalledProcessError, OSError), e:
        progress.update('moving blog data failed, retrying')
        log.info('Could not move data for blog %s: %s', blog_id, e)
        return None

    progress.update('moved blog data')
    log.info('Moved data for blog %s', blog_id)
    return 'move_db'


def process_move_db(blog_id, value, progress):
    progress.update('moving blog database')
    log.info('moving database for %s', blog_id)
    old_mysql_opts = '-h%s -P%s -u%s -p%s' % (value['old_host'],
                                              value['old_port'],
                                              value['old_user'],
                                              value['old_pass'])
    new_mysql_opts = '-h%s -P%s -u%s -p%s' % (value['new_host'],
                                              value['new_port'],
                                              value['new_user'],
                                              value['new_pass'])
    find_tables = ['mysql', old_mysql_opts, value['old_db'], '-NBe',
                   '\\"show tables like \'wp\\_%s\\_%%\'\\"' % blog_id]
    cmd = ['ssh', 'root@%s' % value['old_host'], '"%s"' % ' '.join(find_tables)]
    log.debug('Calling %s', ' '.join(cmd))
    try:
        raw_tables = subprocess.check_output(cmd)
    except (subprocess.CalledProcessError, OSError), e:
        progress.update('could not look up blog tables, retrying')
        log.info('Could not look up tables for blog %s: %s', blog_id, e)
        return None
    tables = ' '.join(t.strip() for t in raw_tables.splitlines())
    move_db = ['mysqldump', '--opt', old_mysql_opts, value['old_db'], tables,
               '|', 'mysql', new_mysql_opts, value['new_db']]
    cmd = ['ssh', 'root@%s' % value['old_host'], '"%s"' % ' '.join(move_db)]
    log.debug('Calling %s', ' '.join(cmd))
    try:
        raw_tables = subprocess.check_output(cmd)
    except (subprocess.CalledProcessError, OSError), e:
        progress.update('could not move blog database, retrying')
        log.info('Could not move blog %s database: %s', blog_id, e)
        return None

    return 'rm'


def process_rm(blog_id, value, progress):
    progress.update('removing blog in old location')
    log.info('removing data of blog %s from %s', blog_id,
             value['old_server_id'])
    return state.STATE_DONE


class NoblogsStateMachine(state.StateMachine):

    states = {
        'error': state.nop,
        'init': init,
        'move_data': process_move_data,
        'move_db': process_move_db,
        'rm': process_rm,
        'done': state.nop,
        }


def main():
    parser = optparse.OptionParser(usage='''%prog [<OPTIONS>]

Noblogs rebalancing driver. Given a set of blog diffs on standard
input, it will perform the necessary migration tasks to ensure the
system reaches the desired state. An HTTP server allows the user to
monitor the migration progress.

The program keeps its state checkpointed to a local database, so
that its execution can be resumed if interrupted. In this case,
rather than passing the data again on standard input, one should
use the `--recover' option.

''')
    parser.add_option('--state', dest='state_file',
                      default='noblogsmv.state',
                      help='Location of the state file (default: %default)')
    parser.add_option('--port', dest='port', type='int',
                      default=3030,
                      help='Port for the web interface (default: %default)')
    parser.add_option('--num-workers', dest='num_workers', type='int',
                      default=20,
                      help='Number of workers (default: %default)')
    parser.add_option('--recover', dest='recover', action='store_true',
                      help='Recover a previous run')
    parser.add_option('--noexit', action='store_true',
                      help='Do not exit when done')
    parser.add_option('--debug', action='store_true')
    opts, args = parser.parse_args()

    if len(args) != 0:
        parser.error('Wrong number of args')

    logging.basicConfig(
        level=(logging.DEBUG if opts.debug else logging.INFO))

    inputfd = None
    if not opts.recover:
        inputfd = sys.stdin
    sm, httpserver = NoblogsStateMachine.new_with_fd(
        opts.state_file,
        inputfd=inputfd,
        num_workers=opts.num_workers,
        enable_http=True,
        http_port=opts.port)

    sm.run()

    if opts.noexit:
        # This will never return.
        httpserver.join()


if __name__ == '__main__':
    main()
