import contextlib
import json
import logging
import os
import sys
import threading
import time
import Queue

import sqlalchemy as sa
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.types import TypeDecorator, VARCHAR
from sqlalchemy.ext.mutable import Mutable

Base = declarative_base()

log = logging.getLogger(__name__)

nop = lambda k, v, progress: v

# Fixed (not application-defined) states.
STATE_INIT = 'init'
STATE_DONE = 'done'
STATE_ERROR = 'error'

EXIT_STATES = (STATE_DONE, STATE_ERROR)


# Encode the data field as JSON
class JSONEncodedDict(TypeDecorator):
    "Represents an immutable structure as a json-encoded string."

    impl = VARCHAR

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            value = json.loads(value)
        return value


class MutableDict(Mutable, dict):

    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()


# The base work unit has a primary key, a state, and some opaque data.
class WorkUnit(Base):

    __tablename__ = 'work'
    key = sa.Column(sa.String(128), primary_key=True, unique=True)
    state = sa.Column(sa.String(6), index=True)
    data = sa.Column(MutableDict.as_mutable(JSONEncodedDict))

    def __init__(self, key, state, data):
        self.key = key
        self.state = state
        self.data = data

    def __eq__(self, b):
        return ((self.key == b.key) and
                (self.state == b.state) and
                (self.data == b.data))


class StateDatabase(object):
    """Simple key-value database that encodes data with a codec."""

    def __init__(self, path, codec=json):
        self.path = path
        self.codec = codec

        self.engine = sa.create_engine('sqlite:///' + path)
        self.session = sessionmaker(bind=self.engine)
        Base.metadata.create_all(self.engine)

    @property
    def conn(self):
        if not hasattr(self._local, 'conn'):
            self._local.conn = sqlite3.connect(self.path)
        return self._local.conn

    def close(self):
        pass

    def put(self, session, work):
        session.add(work)

    def set(self, session, work):
        # The PickleField does not detect changes in the contained
        # object, so we have to create a new one and add that instead.
        #tmp = WorkUnit(work.key, work.state, work.data)
        #self.put(session, tmp)
        work.data = work.data
        self.put(session, work)

    def get(self, session, key):
        return session.query(WorkUnit).get(key)

    def scan(self, session, only_pending=True):
        """Iterate over all the database keys."""
        q = session.query(WorkUnit)
        if only_pending:
            q = q.filter(~WorkUnit.state.in_(EXIT_STATES))
        return (x.key for x in q)

    def dump(self, session):
        """Return every object in the db."""
        return session.query(WorkUnit)

    def are_we_done(self, session):
        """Check if all the entries are in a final state."""
        n = session.query(WorkUnit).filter(~WorkUnit.state.in_(EXIT_STATES)).count()
        return n == 0

    def count_by_state(self, session):
        return dict(session.query('state', 'c').from_statement(
            'select state, count(*) as c from work group by state').all())


@contextlib.contextmanager
def transaction(db):
    session = db.session()
    try:
        yield session
        session.commit()
    except:
        session.rollback()
        raise
    finally:
        session.close()


@contextlib.contextmanager
def work_transaction(db, key):
    with transaction(db) as session:
        work = db.get(session, key)
        yield work
        db.put(session, work)


class WorkerProgressReporter(object):

    def __init__(self, worker):
        self.worker = worker

    def update(self, message):
        self.worker.update_info({'progress': message})


class StateMachineWorker(threading.Thread):
    """A worker.

    The worker algorithm is very simple: it reads one work unit from
    the incoming queue, processes it according to the state definition
    table, and eventually re-injects it into the queue if the result
    state is not final.

    Each worker maintains an 'info' dictionary with some data on the
    current state.
    """

    def __init__(self, sm, input_queue, thread_id):
        threading.Thread.__init__(self, name='Worker %d' % thread_id)
        self.sm = sm
        self.input_queue = input_queue
        self.thread_id = thread_id
        self._info_lock = threading.Lock()
        self._info = {'state': 'init'}

    def run(self):
        log.debug('worker %d started', self.thread_id)
        while True:
            self.set_info({'state': 'idle'})

            key = self.input_queue.get()
            if key is None:
                break

            log.debug('worker %d received key "%s"', self.thread_id, key)
            with work_transaction(self.sm.db, key) as work:
                self.set_info({'state': 'running',
                               'key': key,
                               'value': work.data,
                               'started_at': time.time()})
                progress = WorkerProgressReporter(self)
                new_state = self.sm.process(key, work.state, work.data, progress)
                log.debug('worker %d processed "%s": %s -> %s',
                          self.thread_id, key, work.state, new_state)
                work.state = new_state

            # Reinject if not done.
            if new_state not in EXIT_STATES:
                self.input_queue.put(key)

        log.debug('worker %d exiting', self.thread_id)

    def set_info(self, info):
        with self._info_lock:
            self._info = info

    def update_info(self, d):
        with self._info_lock:
            self._info.update(d)

    @property
    def info(self):
        with self._info_lock:
            info = dict(self._info)
        info['thread_id'] = self.name
        return info


class StatsThread(threading.Thread):

    def __init__(self, sm):
        threading.Thread.__init__(self)
        self.sm = sm
        self._stop = threading.Event()

    def stop(self):
        self._stop.set()

    def run(self):
        while not self._stop.isSet():
            self._stop.wait(10)
            self.sm.compute_stats()


class StateMachine(object):

    # Subclasses should override this, retaining the 'error' rule.
    states = {
        'error': nop,
        }

    def __init__(self, num_workers, dbfile):
        self.db = StateDatabase(dbfile)
        self.running = False
        self.num_workers = num_workers
        self.threads = []
        self.state_count = {}

    def load_data(self, input_stream):
        with transaction(self.db) as session:
            for key, value in input_stream:
                self.db.put(session, WorkUnit(key, STATE_INIT, value))

    def process(self, key, state, value, progress):
        value.pop('error_msg', '')

        method = self.states.get(state)
        if method is None:
            log.error('unknown state "%s" while processing %s: %s',
                      state, key, str(value))
            state = 'error'
        else:
            try:
                new_state = method(key, value, progress)
                if new_state:
                    state = new_state
            except Exception, e:
                log.exception('exception while processing %s: %s',
                              key, str(value))
                progress.update(str(e))
                value['error_msg'] = str(e)
                state = 'error'

        value['count'] = value.get('count', 0) + 1
        return state

    def get_worker_info(self):
        return [t.info for t in self.threads]

    def get_state(self):
        with transaction(self.db) as session:
            return self.db.dump(session)

    def compute_stats(self):
        with transaction(self.db) as session:
            self.state_count = self.db.count_by_state(session)

    def run(self):
        input_queue = Queue.Queue()

        self.threads = [
            StateMachineWorker(self, input_queue, i + 1)
            for i in xrange(self.num_workers)]
        [x.start() for x in self.threads]
        stats_thread = StatsThread(self)
        stats_thread.start()
        self.running = True

        # Inject initial state.
        with transaction(self.db) as session:
            for key in self.db.scan(session):
                input_queue.put(key)

        # Wait until everything is done.
        while True:
            with transaction(self.db) as session:
                if self.db.are_we_done(session):
                    break
            time.sleep(3)

        # Kill the workers.
        for i in xrange(self.num_workers):
            input_queue.put(None)
        [x.join() for x in self.threads]
        stats_thread.stop()
        stats_thread.join()
        self.running = False

    @classmethod
    def new_with_fd(cls, dbfile, inputfd=None, num_workers=10,
                    enable_http=False, http_port=3030):
        sm = cls(num_workers, dbfile)

        if enable_http:
            from noblogsmv import webapp
            httpserver = webapp.WsgiThread(sm, http_port)
            httpserver.start()

        if inputfd:
            def _read_values():
                for line in inputfd:
                    key, value = line.strip().split(' ', 1)
                    yield key, json.loads(value)
            sm.load_data(_read_values())

        if enable_http:
            return sm, httpserver

        return sm


def write_work_unit(fd, key, data):
    fd.write('%s %s\n', key, json.dumps(data))
